"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.empty = [];
/**
 * Build a sort comparator suitable for use with arrays of objects.
 *
 * @param key return the sortable value for an object.
 */
function sortComparatorDesc(key) {
    return (a, b) => {
        const aKey = key(a);
        const bKey = key(b);
        return aKey < bKey ? 1 : aKey > bKey ? -1 : 0;
    };
}
exports.sortComparatorDesc = sortComparatorDesc;
/**
 * Build a sort comparator suitable for use with arrays of objects.
 *
 * @param key return the sortable value for an object.
 */
function sortComparatorAsc(key) {
    return (a, b) => {
        const aKey = key(a);
        const bKey = key(b);
        return aKey > bKey ? 1 : aKey < bKey ? -1 : 0;
    };
}
exports.sortComparatorAsc = sortComparatorAsc;
function sorted(array, key, makeComparator = sortComparatorDesc) {
    return [...array].sort(makeComparator(key));
}
exports.sorted = sorted;
function omitIndex(array, index) {
    return [...array.slice(0, index), ...array.slice(index + 1)];
}
exports.omitIndex = omitIndex;
function omitOne(array, predicate) {
    for (let i = 0; i < array.length; i++) {
        if (predicate(array[i])) {
            return omitIndex(array, i);
        }
    }
    return array;
}
exports.omitOne = omitOne;
function omitNulls(items) {
    return items.reduce(reduceOmitNulls, exports.empty);
}
exports.omitNulls = omitNulls;
function reduceOmitNulls(accum, curr) {
    return curr == null ? accum : [...accum, curr];
}
exports.reduceOmitNulls = reduceOmitNulls;
/**
 * Returns an array with an item appended, if it is not already the last item in the array.
 */
function appendedIfNotAlready(array, item) {
    return array[array.length - 1] !== item ? [...array, item] : array;
}
exports.appendedIfNotAlready = appendedIfNotAlready;
/**
 * Build a filter function that filters an array of items based on a property of
 * each item.
 *
 * Example:
 *
 *     const items = [
 *       { id: 'a', value: 'foo' },
 *       { id: 'a', value: 'bar' },
 *       { id: 'b', value: 'baz' },
 *     ];
 *     items.filter(filterUnique(item => item.id))
 *     // returns [{ id: 'a', value: 'foo' }, { id: 'b', value: 'baz' }]
 */
function filterUnique(key) {
    const seen = new Set();
    return (item) => {
        const itemKey = key(item);
        if (seen.has(itemKey)) {
            return false;
        }
        seen.add(itemKey);
        return true;
    };
}
exports.filterUnique = filterUnique;
/**
 * Return an array of keys, given an object.
 */
function keys(hash) {
    return Object.keys(hash);
}
exports.keys = keys;
function compare(a, b, comparators) {
    for (const comparator of comparators) {
        switch (comparator(a, b)) {
            case 0:
                continue;
            case -1:
                return -1;
            case 1:
                return 1;
        }
    }
    return 0;
}
function sortedIndex(array, element, comparators, start = 0, end = array.length) {
    if (array.length === 0)
        return 0;
    const pivot = (start + end) >> 1; // should be faster than dividing by 2
    const c = compare(element, array[pivot], comparators);
    if (end - start <= 1)
        return c == -1 ? pivot : pivot + 1;
    switch (c) {
        case -1:
            return sortedIndex(array, element, comparators, start, pivot);
        case 0:
            return pivot;
        case 1:
            return sortedIndex(array, element, comparators, pivot, end);
    }
}
exports.sortedIndex = sortedIndex;
function sortedInsert(sortedArray, item, comparators) {
    const index = sortedIndex(sortedArray, item, comparators);
    return [...sortedArray.slice(0, index), item, ...sortedArray.slice(index)];
}
exports.sortedInsert = sortedInsert;
function get(array, index, outOfBoundsValue) {
    return index >= array.length ? outOfBoundsValue : array[index];
}
exports.get = get;
function repeat(item, count) {
    const arr = [];
    for (let i = 0; i < count; i++) {
        arr.push(item);
    }
    return arr;
}
exports.repeat = repeat;
function zeroes(n) {
    return repeat(0, n);
}
exports.zeroes = zeroes;
//# sourceMappingURL=index.js.map